@@ -17,7 +17,11 @@ try:
 except ImportError:
     datetime = None
 
-from utils import threadeddict, storage, iters, iterbetter
+try: set
+except NameError:
+    from sets import Set as set
+    
+from utils import threadeddict, storage, iters, iterbetter, safestr, safeunicode
 
 try:
     # db module can work independent of web.py
@@ -102,7 +106,7 @@ class SQLQuery:
     """
     # tested in sqlquote's docstring
     def __init__(self, items=[]):
-        """Creates a new SQLQuery.
+        r"""Creates a new SQLQuery.
         
             >>> SQLQuery("x")
             <sql: 'x'>
@@ -121,7 +125,7 @@ class SQLQuery:
         elif isinstance(items, SQLQuery):
             self.items = list(items.items)
         else:
-            self.items = [str(items)]
+            self.items = [items]
             
         # Take care of SQLLiterals
         for i, item in enumerate(self.items):
@@ -167,12 +171,20 @@ class SQLQuery:
             >>> q.query(paramstyle='qmark')
             'SELECT * FROM test WHERE name=?'
         """
-        s = ''
+        s = []
         for x in self.items:
             if isinstance(x, SQLParam):
                 x = x.get_marker(paramstyle)
-            s += x
-        return s
+                s.append(safestr(x))
+            else:
+                x = safestr(x)
+                # automatically escape % characters in the query
+                # For backward compatability, ignore escaping when the query looks already escaped
+                if paramstyle in ['format', 'pyformat']:
+                    if '%' in x and '%%' not in x:
+                        x = x.replace('%', '%%')
+                s.append(x)
+        return "".join(s)
     
     def values(self):
         """
@@ -200,12 +212,18 @@ class SQLQuery:
         return q
     
     join = staticmethod(join)
-
-    def __str__(self):
+    
+    def _str(self):
         try:
-            return self.query() % tuple([sqlify(x) for x in self.values()])
+            return self.query() % tuple([sqlify(x) for x in self.values()])            
         except (ValueError, TypeError):
             return self.query()
+        
+    def __str__(self):
+        return safestr(self._str())
+        
+    def __unicode__(self):
+        return safeunicode(self._str())
 
     def __repr__(self):
         return '<sql: %s>' % repr(str(self))
@@ -227,6 +245,20 @@ class SQLLiteral:
 
 sqlliteral = SQLLiteral
 
+def _sqllist(values):
+    """
+        >>> _sqllist([1, 2, 3])
+        <sql: '(1, 2, 3)'>
+    """
+    items = []
+    items.append('(')
+    for i, v in enumerate(values):
+        if i != 0:
+            items.append(', ')
+        items.append(sqlparam(v))
+    items.append(')')
+    return SQLQuery(items)
+
 def reparam(string_, dictionary): 
     """
     Takes a string and a dictionary and interpolates the string
@@ -234,6 +266,8 @@ def reparam(string_, dictionary):
 
         >>> reparam("s = $s", dict(s=True))
         <sql: "s = 't'">
+        >>> reparam("s IN $s", dict(s=[1, 2]))
+        <sql: 's IN (1, 2)'>
     """
     dictionary = dictionary.copy() # eval mucks with it
     vals = []
@@ -241,7 +275,7 @@ def reparam(string_, dictionary):
     for live, chunk in _interpolate(string_):
         if live:
             v = eval(chunk, dictionary)
-            result.append(sqlparam(v))
+            result.append(sqlquote(v))
         else: 
             result.append(chunk)
     return SQLQuery.join(result, '')
@@ -269,6 +303,7 @@ def sqlify(obj):
     elif datetime and isinstance(obj, datetime.datetime):
         return repr(obj.isoformat())
     else:
+        if isinstance(obj, unicode): obj = obj.encode('utf8')
         return repr(obj)
 
 def sqllist(lst): 
@@ -338,8 +373,13 @@ def sqlquote(a):
 
         >>> 'WHERE x = ' + sqlquote(True) + ' AND y = ' + sqlquote(3)
         <sql: "WHERE x = 't' AND y = 3">
+        >>> 'WHERE x = ' + sqlquote(True) + ' AND y IN ' + sqlquote([2, 3])
+        <sql: "WHERE x = 't' AND y IN (2, 3)">
     """
-    return sqlparam(a).sqlquery()
+    if isinstance(a, list):
+        return _sqllist(a)
+    else:
+        return sqlparam(a).sqlquery()
 
 class Transaction:
     """Database transaction."""
@@ -414,8 +454,13 @@ class DB:
     def __init__(self, db_module, keywords):
         """Creates a database.
         """
+        # some DB implementaions take optional paramater `driver` to use a specific driver modue
+        # but it should not be passed to connect
+        keywords.pop('driver', None)
+
         self.db_module = db_module
         self.keywords = keywords
+
         
         self._ctx = threadeddict()
         # flag to enable/disable printing queries
@@ -507,17 +552,6 @@ class DB:
             return '%s'
         raise UnknownParamstyle, style
 
-    def _py2sql(self, val):
-        """
-        Transforms a Python value into a value to pass to cursor.execute.
-
-        This exists specifically for a workaround in SqliteDB.
-
-        """
-        if isinstance(val, unicode):
-            val = val.encode('UTF-8')
-        return val
-
     def _db_execute(self, cur, sql_query): 
         """executes an sql query"""
         self.ctx.dbq_count += 1
@@ -525,9 +559,7 @@ class DB:
         try:
             a = time.time()
             paramstyle = getattr(self, 'paramstyle', 'pyformat')
-            out = cur.execute(sql_query.query(paramstyle),
-                              [self._py2sql(x)
-                               for x in sql_query.values()])
+            out = cur.execute(sql_query.query(paramstyle), sql_query.values())
             b = time.time()
         except:
             if self.printing:
@@ -845,32 +877,41 @@ class PostgresDB(DB):
     """Postgres driver."""
     def __init__(self, **keywords):
         if 'pw' in keywords:
-            keywords['password'] = keywords['pw']
-            del keywords['pw']
+            keywords['password'] = keywords.pop('pw')
             
-        db_module = self.get_db_module()
-        keywords['database'] = keywords.pop('db')
+        db_module = import_driver(["psycopg2", "psycopg", "pgdb"], preferred=keywords.pop('driver', None))
+        if db_module.__name__ == "psycopg2":
+            import psycopg2.extensions
+            psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
+
+        # if db is not provided postgres driver will take it from PGDATABASE environment variable
+        if 'db' in keywords:
+            keywords['database'] = keywords.pop('db')
+        
         self.dbname = "postgres"
         self.paramstyle = db_module.paramstyle
         DB.__init__(self, db_module, keywords)
         self.supports_multiple_insert = True
+        self._sequences = None
         
-    def get_db_module(self):
-        try: 
-            import psycopg2 as db
-            import psycopg2.extensions
-            psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
-        except ImportError: 
-            try: 
-                import psycopg as db
-            except ImportError: 
-                import pgdb as db
-        return db
-
     def _process_insert_query(self, query, tablename, seqname):
-        if seqname is None: 
+        if seqname is None:
+            # when seqname is not provided guess the seqname and make sure it exists
             seqname = tablename + "_id_seq"
-        return query + "; SELECT currval('%s')" % seqname
+            if seqname not in self._get_all_sequences():
+                seqname = None
+        
+        if seqname:
+            query += "; SELECT currval('%s')" % seqname
+            
+        return query
+    
+    def _get_all_sequences(self):
+        """Query postgres to find names of all sequences used in this database."""
+        if self._sequences is None:
+            q = "SELECT c.relname FROM pg_class c WHERE c.relkind = 'S'"
+            self._sequences = set([c.relname for c in self.query(q)])
+        return self._sequences
 
     def _connect(self, keywords):
         conn = DB._connect(self, keywords)
@@ -902,17 +943,26 @@ class MySQLDB(DB):
     def _process_insert_query(self, query, tablename, seqname):
         return query, SQLQuery('SELECT last_insert_id();')
 
+def import_driver(drivers, preferred=None):
+    """Import the first available driver or preferred driver.
+    """
+    if preferred:
+        drivers = [preferred]
+
+    for d in drivers:
+        try:
+            return __import__(d, None, None, ['x'])
+        except ImportError:
+            pass
+    raise ImportError("Unable to import " + " or ".join(drivers))
+
 class SqliteDB(DB): 
     def __init__(self, **keywords):
-        try:
-            import sqlite3 as db
+        db = import_driver(["sqlite3", "pysqlite2.dbapi2", "sqlite"], preferred=keywords.pop('driver', None))
+
+        if db.__name__ in ["sqlite3", "pysqlite2.dbapi2"]:
             db.paramstyle = 'qmark'
-        except ImportError:
-            try:
-                from pysqlite2 import dbapi2 as db
-                db.paramstyle = 'qmark'
-            except ImportError:
-                import sqlite as db
+
         self.paramstyle = db.paramstyle
         keywords['database'] = keywords.pop('db')
         self.dbname = "sqlite"        
@@ -924,33 +974,9 @@ class SqliteDB(DB):
     def query(self, *a, **kw):
         out = DB.query(self, *a, **kw)
         if isinstance(out, iterbetter):
-            # rowcount is not provided by sqlite
             del out.__len__
         return out
 
-    # as with PostgresDB, the database is assumed to be in UTF-8.
-    # This doesn't mean we turn byte-strings coming out of it into
-    # Unicode objects, but we avoid trying to put Unicode objects into
-    # it.
-    encoding = 'UTF-8'
-
-    def _py2sql(self, val):
-        r"""
-        Work around a couple of problems in SQLite that maybe pysqlite
-        should take care of: give it True and False and it thinks
-        they're column names; give it Unicode and it tries to insert
-        it in, possibly, ASCII.
-
-            >>> meth = SqliteDB(db='nonexistent')._py2sql
-            >>> [meth(x) for x in [True, False, 1, 2, 'foo', u'souffl\xe9']]
-            [1, 0, 1, 2, 'foo', 'souffl\xc3\xa9']
-
-        """
-        if val is True: return 1
-        elif val is False: return 0
-        elif isinstance(val, unicode): return val.encode(self.encoding)
-        else: return val
-
 class FirebirdDB(DB):
     """Firebird Database.
     """
@@ -988,11 +1014,35 @@ class MSSQLDB(DB):
     def __init__(self, **keywords):
         import pymssql as db    
         if 'pw' in keywords:
-            keywords['password'] = keywords.pop('kw')
+            keywords['password'] = keywords.pop('pw')
         keywords['database'] = keywords.pop('db')
         self.dbname = "mssql"
         DB.__init__(self, db, keywords)
 
+    def sql_clauses(self, what, tables, where, group, order, limit, offset): 
+        return (
+            ('SELECT', what),
+            ('TOP', limit),
+            ('FROM', sqllist(tables)),
+            ('WHERE', where),
+            ('GROUP BY', group),
+            ('ORDER BY', order),
+            ('OFFSET', offset))
+            
+    def _test(self):
+        """Test LIMIT.
+
+            Fake presence of pymssql module for running tests.
+            >>> import sys
+            >>> sys.modules['pymssql'] = sys.modules['sys']
+            
+            MSSQL has TOP clause instead of LIMIT clause.
+            >>> db = MSSQLDB(db='test', user='joe', pw='secret')
+            >>> db.select('foo', limit=4, _test=True)
+            <sql: 'SELECT * TOP 4 FROM foo'>
+        """
+        pass
+
 class OracleDB(DB): 
     def __init__(self, **keywords): 
         import cx_Oracle as db 
